@@ -1,6 +1,8 @@
 # MANAGER, WORKER, and NODE can send
 DATA_ACK = "DATA_ACK"
 SHUTDOWN = "SHUTDOWN"
+FILE_TRANSFER_START = "FILE_TRANSFER_START"
+FILE_TRANSFER_END = "FILE_TRANSFER_END"
 
 # SEND BY MANAGER
 MANAGER_MESSAGE = "MANAGER_MESSAGE"
@@ -12,6 +14,7 @@ MANAGER_REQUEST_STEP = "MANAGER_REQUEST_STEP"
 MANAGER_START_NODES = "MANAGER_START_NODES"
 MANAGER_STOP_NODES = "MANAGER_STOP_NODES"
 MANAGER_REQUEST_GATHER = "MANAGER_REQUEST_GATHER"
+MANAGER_REQUEST_COLLECT = "MANAGER_REQUEST_COLLECT"
 
 # SEND BY WORKER
 WORKER_MESSAGE = "WORKER_MESSAGE"
@@ -26,6 +29,7 @@ WORKER_REQUEST_GATHER = "WORKER_REQUEST_GATHER"
 WORKER_REPORT_GATHER = "WORKER_REPORT_GATHER"
 WORKER_START_NODES = "WORKER_START_NODES"
 WORKER_STOP_NODES = "WORKER_STOP_NODES"
+WORKER_TRANSFER_COMPLETE = "WORKER_TRANSFER_COMPLETE"
 
 # SEND BY NODE
 NODE_MESSAGE = "NODE_MESSAGE"
